// hlavickove subory *******************************************************
#include "common/includes.h"
// vlastny .h subor
#include "backend.h"

// nacitanie 3ds suborov
#include "common/load3ds.h"

CLoad3ds *l3ds; // instancia triedy CLoad3ds - praca s 3ds subormi

CMath3d m3d; // instancia triedy CMath3d - praca s matematikou

// zoznam vsetkych trojuholnikov
vector<_tri> tris;
// zoznam vsetkych svetelnych zdrojov
vector<_light> lights;

// konstruktor *************************************************************
CBackend::CBackend()
{
 l3ds=NULL;

 // reset zoznamu vsetkych trojuholnikov
 tris.reset();
 // reset zoznamu vsetkych svetelnych zdrojov
 lights.reset();
}

// destruktor **************************************************************
CBackend::~CBackend()
{

}

// nacitanie vsetkych trojuholnikov ****************************************
int CBackend::tris_add(void)
{
 // vytvorenie instancie triedy CLoad3ds
 l3ds=new CLoad3ds;
 if (l3ds==NULL) {
  printf("Nemozno vytvorit instanciu triedy CLoad3ds!\n"); return(1);
 }

 // nacitanie obsahu 3ds suboru
 if (l3ds->load("data/models/scene.3ds")) {
  DELETEP(l3ds); return(1);
 }

 // ziskanie trojuholnikov patriacich objektom
 // ziskanie informacii o svetelnych zdrojoch definovanych v 3ds subore
 l3ds->get_tris_and_lights();

 // obsah 3ds suboru nie je potrebne dalej uchovavat
 DELETEP(l3ds);

 // doplnenie udajov nevyhnutnych pre pracu s trojuholnikmi
 tris_upgrade();

 // vytvorenie lightmap pre vsetky trojuholniky
 tris_lightmaps();

 return(0);
}

// vykreslenie vsetkych trojuholnikov **************************************
void CBackend::tris_draw(void)
{
 // vykreslenie trojuholnikov nacitanych z 3ds suboru
 if (l3ds!=NULL) l3ds->forcedraw();
}

// doplnenie udajov nevyhnutnych pre pracu s trojuholnikmi *****************
void CBackend::tris_upgrade(void)
{
 // pre vsetky trojuholniky
 loopk(tris.length()) {
  // usporiadanie vrcholov trojuholnika na zaklade prepony
  // vsetky trojuholniky budu mat vrcholy usporiadane tak,
  // aby bola prepona vzdy tvorena vrcholmi 2 - 0
  // pozadovane poradie vrcholov:
  //         2         0 - 1, 1 - 2 - odvesny trojuholnika
  //       . .         2 - 0        - prepona trojuholnika
  //     .   .
  //   .     .
  // 0 ..... 1
  a0=tris[k].v[1]-tris[k].v[0]; // vektor 0 - 1
  a1=tris[k].v[2]-tris[k].v[1]; // vektor 1 - 2
  a2=tris[k].v[0]-tris[k].v[2]; // vektor 2 - 0
  m0=a0.magnitude(); // dlzky vektorov
  m1=a1.magnitude();
  m2=a2.magnitude();
  // ulozenie zdrojovych udajov
  b0=tris[k].t[0]; b1=tris[k].t[1]; b2=tris[k].t[2];
  a0=tris[k].v[0]; a1=tris[k].v[1]; a2=tris[k].v[2];
  // ak je najldhsia strana 0 - 1, vrcholy sa usporiadaju
  if ((m0>m1)&&(m0>m2)) {
   tris[k].t[0]=b1; tris[k].t[1]=b2; tris[k].t[2]=b0;    // texturovacie suradnice
   tris[k].v[0]=a1; tris[k].v[1]=a2; tris[k].v[2]=a0;    // suradnice vrcholov
   a0=tris[k].vn[0]; a1=tris[k].vn[1]; a2=tris[k].vn[2];
   tris[k].vn[0]=a1; tris[k].vn[1]=a2; tris[k].vn[2]=a0; // normaly vo vrcholoch
  }    // ak je najdlhsia strana 1 - 2, vrcholy sa usporiadaju
  else if ((m1>m0)&&(m1>m2)) {
        tris[k].t[0]=b2; tris[k].t[1]=b0; tris[k].t[2]=b1;
        tris[k].v[0]=a2; tris[k].v[1]=a0; tris[k].v[2]=a1;
        a0=tris[k].vn[0]; a1=tris[k].vn[1]; a2=tris[k].vn[2];
        tris[k].vn[0]=a2; tris[k].vn[1]=a0; tris[k].vn[2]=a1;
       }
  // ak je najdlhsia strana 2 - 0, poradie vrcholov sa nemeni

  // vypocet normaly trojuholnika
  a0=tris[k].v[1]-tris[k].v[0]; a1=tris[k].v[2]-tris[k].v[0];
  // vektorovy sucin dvoch vektorov = normala trojuholnika
  a2=a0.cross(a1); a2.normalize();
  tris[k].n=a2;
  // vypocet prvku D rovnice roviny, ktoru tvori trojuholnik
  // D=-(AX+BY+CZ)
  // A,B,C - normala trojuholnika
  // X,Y,Z - lubovolny vrchol trojuholnika
  tris[k].d=-(a2.x*tris[k].v[0].x+a2.y*tris[k].v[0].y+a2.z*tris[k].v[0].z);

  // vypocet koliznych normal
  a1=tris[k].n; // normala trojuholnika
  loopi(3) {
   if (i==0) a0=tris[k].v[1]-tris[k].v[0]; // 1. strana trojuholnika
   if (i==1) a0=tris[k].v[2]-tris[k].v[1]; // 2. strana trojuholnika
   if (i==2) a0=tris[k].v[0]-tris[k].v[2]; // 3. strana trojuholnika
   // vektorovy sucin vektorov = kolizna normala
   // prvky koliznej normaly su prvkami rovnice roviny
   // A,B,C, nasledne sa vypocita prvok D
   a2=a0.cross(a1); a2.normalize();
   tris[k].cn[i].x=a2.x; tris[k].cn[i].y=a2.y; tris[k].cn[i].z=a2.z;
   // z rovnice roviny AX+BY+CZ+D=0 => D=-(AX+BY+CZ)
   tris[k].cn[i].w=-(a2.x*tris[k].v[i].x+a2.y*tris[k].v[i].y+a2.z*tris[k].v[i].z);
  }
 }
}

// reset indikatora spracovania pre vsetky trojuholniky ********************
void CBackend::tris_done(void)
{
 loopk(tris.length()) tris[k].done=false;
}

// vytvorenie lightmap pre vsetky trojuholniky *****************************
int CBackend::tris_lightmaps(void)
{
 GLint quadTri; // identifikator trojuholnika, ktory spolu s aktualnym
                // trojuholnikom vytvara stvoruholnik

 // pocet segmentov lightmapy v smere X
 // 512/128=4   4*4  =spolu 16 segmentov
 // 512/64 =8   8*8  =spolu 64 segmentov
 // 512/32 =16  16*16=spolu 256 segmentov
 segsNum=LM_SIZE/LM_SEGMENT;

 // INFO: pocet vsetkych trojuholnikov
 printf("Pocet vsetkych trojuholnikov: %u\n",tris.length());

 // aktualna lightmapa, aktualny segment lightmapy
 act_lm=0; act_seg=0;

 // vytvorenie 1. lightmapy
 if (texture.createLightmap(lmid[act_lm])) {
  printf("Chyba vytvorenia lightmapy!\n"); return(1);
 }

 // reset indikatora spracovania pre vsetky trojuholniky
 tris_done();

 // vypocet odtienov jednotlivych lumelov lightmap pre vsetky trojuholniky
 loopk(tris.length()) {

  // ak uz bol trojuholnik spracovany, cyklus pokracuje na dalsom trojuholniku
  if (tris[k].done) continue;
  // trojuholnik sa spracuje => dalej sa uz nebude spracovavat
  tris[k].done=true;
  // priradenie identifikatora lightmapy trojuholniku
  tris[k].lmid=texture.lm_id[lmid[act_lm]];

  // vyber trojuholnika, ktory spolu s aktualnym trojuholnikom vytvara
  // stvoruholnik, taketo trojuholniky musia mat 2 spolocne vrcholy
  // a ich normaly musia byt "vyhovujuco" rovnobezne
  quadTri=-1;
  loopl(tris.length()) {

   // vyber trojuholnika, ktory este nebol spracovany
   if (tris[l].done) continue;
   // ak normaly trojuholnikov nie su "vyhovujuco" rovnobezne,
   // cyklus pokracuje na dalsom trojuholniku
   a0=tris[k].n; a1=tris[l].n; // normaly trojuholnikov
   m0=a0.dot(a1);              // skalarnym sucinom sa urci rovnobeznost
   if (m0<0.9f) continue;
   // normaly trojuholnikov su "vyhovujuco" rovnobezne, zostava
   // zistit, ci maju 2 spolocne vrcholy, pretoze su vrcholy vsetkych
   // trojuholnikov vzdy usporiadane na zaklade prepony, staci
   // porovnat vrcholy 0 a 2
   if ((tris[k].v[0]==tris[l].v[2])&&(tris[k].v[2]==tris[l].v[0])) {
    // trojuholnik sa spracuje => dalej sa uz nebude spracovavat
    tris[l].done=true;
    // priradenie rovnakeho identifikatora lightmapy
    tris[l].lmid=texture.lm_id[act_lm];
    quadTri=l;
    break; // boli najdene 2 trojuholniky vytvarajuce stvoruholnik
   }

  } // pre vsetky trojuholniky - loopl

  // ak bol pre aktualny trojuholnik najdeny dalsi trojuholnik, s ktorym
  // vytvara stvoruholnik, trojuholniky sa spracuju spolocne, v opacnom pripade
  // sa aktualny trojuholnik spracuje samostatne
  // vypocet odtiena lumelov patriacich konkretnemu segmentu lightmapy
  // vyhradenemu pre 1 konkretny trojuholnik, resp. 2 trojuholniky
  // vytvarajuce stvoruholnik
  if (set_lumels(k,quadTri,act_seg)) return(1);

  // vyber nasledujuceho segmentu lightmapy
  act_seg++;
  // po prekroceni definovaneho poctu segmentov bude vytvorena nova lightmapa
  if (act_seg>=(segsNum*segsNum)) {
   // aktualna lightmapa, aktualny segment lightmapy
   act_lm++; act_seg=0;

   // vytvorenie novej lightmapy
   if (texture.createLightmap(lmid[act_lm])) {
    printf("Chyba vytvorenia lightmapy!\n"); return(1);
   }
  }

 } // pre vsetky trojuholniky - loopk

 printf("Pocet vytvorenych lightmap: %u\n",act_lm+1);

 return(0);
}

// vypocet odtiena lumelov patriacich konkretnemu segmentu lightmapy *******
int CBackend::set_lumels(GLint K,GLint L,GLuint N)
{
 GLfloat x1,y1,x2,y2; // hranice segmentu lightmapy
 GLubyte *data;       // pomocna pamat
 bool collision;      // indikator kolizie svetelneho luca s trojuholnikom

 vec3 lumel_pos; // poloha lumelu - bodu patriaceho lightmape
 vec3 light_pos; // poloha svetelneho zdroja

 GLint triID; // identifikator aktualneho trojuholnika

 GLint luminance; // odtien lumelu

 // vyhradenie pamate pre odtien lumelov lightmapy
 data=new GLubyte[LM_SEGMENT*LM_SEGMENT]; // 1 Byte / lumel
 if (data==NULL) { printf("Nemozno vyhradit pamat!\n"); return(1); }
 // nastavenie neutralnej farby lumelov
 memset(data,LM_DIP,LM_SEGMENT*LM_SEGMENT);

 // hranice segmentu lightmapy
 x1=(GLfloat) (N%segsNum)/(GLfloat) segsNum;
 y1=(GLfloat) (N/segsNum)/(GLfloat) segsNum;
 x2=(GLfloat) ((N%segsNum)+1)/(GLfloat) segsNum;
 y2=(GLfloat) ((N/segsNum)+1)/(GLfloat) segsNum;
 // posun hranic segmentu lightmapy smerom dovnutra o 1 lumel,
 // odtien okrajovych lumelov bude doplneny neskor, dovodom
 // je korekcia GL_LINEAR texturovania, ktore zmiesava farby
 // bodov textury pricom vznikaju nechcene prechody medzi
 // odtienami lumelov lightmap dvoch susediacich trojuholnikov
 m0=(x2-x1)/(GLfloat) LM_SEGMENT;
 x1+=m0; y1+=m0; x2-=m0; y2-=m0;
 // priradenie texturovacich suradnic pre lightmapu aktualneho trojuholnika
 tris[K].l[0].x=x1; tris[K].l[0].y=y1; //            x2,y2 l2
 tris[K].l[1].x=x2; tris[K].l[1].y=y1; //              |
 tris[K].l[2].x=x2; tris[K].l[2].y=y2; // x1,y1 l0 - x2,y1 l1
 // rovnaky segment lightmapy sa pouzije aj pre 2. trojuholnik
 if (L!=-1) {
  // priradenie texturovacich suradnic pre lightmapu 2. trojuholnika
  tris[L].l[0].x=x2; tris[L].l[0].y=y2; // x1,y2 l1 - x2,y2 l0
  tris[L].l[1].x=x1; tris[L].l[1].y=y2; //   |
  tris[L].l[2].x=x1; tris[L].l[2].y=y1; // x1,y1 l2
 }

 // vypocet odtiena lumelov
 loopj(LM_SEGMENT-2) { // v smere Y
  loopi(LM_SEGMENT-2) { // v smere X
   // 2D suradnice lumelu
   lumel_pos.x=(x2-x1)*((GLfloat) i+0.5f)/(GLfloat) (LM_SEGMENT-2);
   lumel_pos.y=(y2-y1)*((GLfloat) j+0.5f)/(GLfloat) (LM_SEGMENT-2);

   if (j<=i) { // prava dolna polovica segmentu
    triID=K;
    // 3D suradnice lumelu
    lumel_worldpos(&lumel_pos,tris[K].v[0],tris[K].v[1],tris[K].v[2],
                              tris[K].l[0],tris[K].l[1],tris[K].l[2],
                              tris[K].n,tris[K].d);
   }
   else {      // lava horna polovica segmentu
    triID=L;
    if (L!=-1) {
     // 3D suradnice lumelu
     lumel_worldpos(&lumel_pos,tris[L].v[2],tris[L].v[0],tris[L].v[1],
                               tris[L].l[2],tris[L].l[0],tris[L].l[1],
                               tris[L].n,tris[L].d);
    }
   }

   // ak sa pocita pre lavu hornu polovicu, ktora nie je pridelena
   // ziadnemu trojuholniku, cyklus pokracuje na dalsom lumene
   if (triID==-1) continue;

   // pre vsetky svetelne zdroje
   loopl(lights.length()) {
    // poloha svetelneho zdroja
    light_pos=lights[l].p;
    // vektor poloha svetelneho zdroja -> poloha lumelu
    a0=lumel_pos-light_pos;
    // ak je svetelny zdroj privrateny k trojuholniku
    // v takom pripade moze tento trojuholnik osvetlit
    if (a0.dot(tris[triID].n)>0.0f) continue;

    collision=false;
    // detekcia kolizie luca a vsetkych ostatnych trojuholnikov
    loopk(tris.length()) {
     // detekcia sa vykona iba so vsetkymi ostatnymi trojuholnikmi,
     // nie s tym istym trojuholnikom
     if (tris[k].id!=tris[triID].id) {
      // kolizia svetelneho luca s trojuholnikom
      if (m3d.ray_tri_collision(lumel_pos,light_pos,tris[k].n,tris[k].d,
                                tris[k].cn[0],tris[k].cn[1],tris[k].cn[2])) {
       collision=true;
       break;
      }
     }
    } // pre vsetky ostatne trojuholniky

    // ak nenastala kolizia, lumel je osvetleny
    if (!collision) {
     // vzdialenost lumelu od svetelneho zdroja
     m0=a0.magnitude(); m0*=m0;
     // povodny edtien lumelu
     luminance=data[(i+1)+((j+1)*LM_SEGMENT)];
     // novy odtien lumelu
     luminance+=255-((GLint) m0);
     if (luminance<LM_DIP) luminance=LM_DIP; if (luminance>255) luminance=255;
     // ak je novy odtien svetlejsi ako povodny
     if (luminance>data[(i+1)+((j+1)*LM_SEGMENT)]) data[(i+1)+((j+1)*LM_SEGMENT)]=luminance;
    }

   } // pre vsetky svetelne zdroje

  } // v smere X
 } // v smere Y

 // korekcia okrajovych casti segmentu lightmapy
 // odtien lumelov sa skopiruje zo susednych lumelov
 loopi(LM_SEGMENT) {
  // spodny a vrchny okraj segmentu
  if (data[i+1*LM_SEGMENT]!=LM_DIP) data[i+0*LM_SEGMENT]=data[i+1*LM_SEGMENT];
  if (data[i+(LM_SEGMENT-2)*LM_SEGMENT]!=LM_DIP) data[i+(LM_SEGMENT-1)*LM_SEGMENT]=data[i+(LM_SEGMENT-2)*LM_SEGMENT];
  // lavy a pravy okraj segmentu
  if (data[1+i*LM_SEGMENT]!=LM_DIP) data[0+i*LM_SEGMENT]=data[1+i*LM_SEGMENT];
  if (data[(LM_SEGMENT-2)+i*LM_SEGMENT]!=LM_DIP) data[(LM_SEGMENT-1)+i*LM_SEGMENT]=data[(LM_SEGMENT-2)+i*LM_SEGMENT];
 }
 loopi(LM_SEGMENT-1) { // diagonala segmentu
  if (i==0) continue;
  if ((data[(i+1)+i*LM_SEGMENT]!=LM_DIP)&&(data[i+(i-1)*LM_SEGMENT]!=LM_DIP)) {
   data[i+i*LM_SEGMENT]=(data[(i+1)+i*LM_SEGMENT]+data[i+(i-1)*LM_SEGMENT])/2;
  }
 }
 // rozmazanie okrajov osvetlenych casti
 loopj(LM_SEGMENT-1) {
  loopi(LM_SEGMENT-1) {
   if ((i==0)||(j==0)) continue;
   // vypocita sa priemerna hodnota odtiena 8-mich okolitych lumelov
   data[i+j*LM_SEGMENT]=(data[(i-1)+(j-1)*LM_SEGMENT]+data[(i+0)+(j-1)*LM_SEGMENT]+
                         data[(i+1)+(j-1)*LM_SEGMENT]+data[(i-1)+(j+0)*LM_SEGMENT]+
                         data[(i+1)+(j+0)*LM_SEGMENT]+data[(i-1)+(j+1)*LM_SEGMENT]+
                         data[(i+0)+(j+1)*LM_SEGMENT]+data[(i+1)+(j+1)*LM_SEGMENT])/8;
  }
 }
 // update odtiena lumelov konkretneho segmentu lightmapy
 glBindTexture(GL_TEXTURE_2D,tris[K].lmid);
 glTexSubImage2D(GL_TEXTURE_2D,0,
                 (N%segsNum)*LM_SEGMENT,(N/segsNum)*LM_SEGMENT,LM_SEGMENT,LM_SEGMENT,
                 GL_LUMINANCE,GL_UNSIGNED_BYTE,data);

 // uvolnenie vyhradenej pamate
 DELETEA(data);

 return(0);
}

// vypocet 3D suradnic lumelu **********************************************
void CBackend::lumel_worldpos(vec3 *P,vec3 V0,vec3 V1,vec3 V2,
                              vec2 T0,vec2 T1,vec2 T2,
                              vec3 N,float D)
{
 // pomocne premenne pre vypocet 3D suradnic lumelov
 GLfloat dpdu,dpdv,dqdu,dqdv,den;
 GLfloat aa,bb,cc;

 // ak je niektora zlozka normalny zaporna, je potrebne
 // ju negovat z dovodu korektneho vyberu projekcie do roviny
 a0=N;
 if (a0.x<0.0f) a0.x*=-1.0f;
 if (a0.y<0.0f) a0.y*=-1.0f;
 if (a0.z<0.0f) a0.z*=-1.0f;

 // denominator=(v0-v2)(u1-u2)-(v1-v2)(u0-u2)
 // vypocita sa z texturovacich suradnic U a V
 // pricom U = T.x, V = T.y
 den=(T0.y-T2.y)*(T1.x-T2.x)-(T1.y-T2.y)*(T0.x-T2.x);
 if (den==0.0f) den=1.0f;
 den=1.0f/den; // reciprocna hodnota

 // nasledovne parametre vznikli odvodenim zo vzorcov
 // ziskanych konstrukciou dvoch podobnych trojuholnikov
 // pomocou nich mozno nasledne vypocitat 3D suradnice lumelu,
 // pri tychto vypoctoch dochadza k posunu 3D vrcholu po celej
 // ploche trojuholnika v 3D priestore
 // dpdu=((x1-x2)(v0-v2)-(x0-x2)(v1-v2))/denominator
 // dpdv=((x1-x2)(u0-u2)-(x0-x2)(u1-u2))/-denominator
 // dqdu=((y1-y2)(v0-v2)-(y0-y2)(v1-v2))/denominator
 // dqdv=((y1-y2)(u0-u2)-(y0-y2)(u1-u2))/-denominator

 // projekcia 3D trojuholnika do 2D roviny, na zaklade
 // najvacsieho prvku normaly trojuholnika (roviny)
 // urcime, do ktorej pravouhlej roviny bude trojuholnik
 // projektovany, nasledne z rovnice roviny AX+BY+CZ+D=0
 // ziskame 3. suradnicu:
 // x=-(BY+CZ+D)/A y=-(AX+CZ+D)/B z=-(AX+BY+D)/C
 // pos0=v0.0+(dpdu*LP->x)+(dpdv*LP->y);
 // pos1=v0.1+(dqdu*LP->x)+(dqdv*LP->y);

 // 3D suradnice lumelu
 if ((a0.x>=a0.y)&&(a0.x>=a0.z)) {      // projekcia do roviny YZ
  if (N.x==0.0f) N.x=1.0f;

  dpdu= den*((V1.y-V2.y)*(T0.y-T2.y)-(V0.y-V2.y)*(T1.y-T2.y));
  dpdv=-den*((V1.y-V2.y)*(T0.x-T2.x)-(V0.y-V2.y)*(T1.x-T2.x));
  dqdu= den*((V1.z-V2.z)*(T0.y-T2.y)-(V0.z-V2.z)*(T1.y-T2.y));
  dqdv=-den*((V1.z-V2.z)*(T0.x-T2.x)-(V0.z-V2.z)*(T1.x-T2.x));

  bb=V0.y+(dpdu*P->x)+(dpdv*P->y);
  cc=V0.z+(dqdu*P->x)+(dqdv*P->y);
  aa=-(N.y*bb+N.z*cc+D)/N.x;
 }
 else if ((a0.y>=a0.x)&&(a0.y>=a0.z)) { // projekcia do roviny XZ
       if (N.y==0.0f) N.y=1.0f;

       dpdu= den*((V1.x-V2.x)*(T0.y-T2.y)-(V0.x-V2.x)*(T1.y-T2.y));
       dpdv=-den*((V1.x-V2.x)*(T0.x-T2.x)-(V0.x-V2.x)*(T1.x-T2.x));
       dqdu= den*((V1.z-V2.z)*(T0.y-T2.y)-(V0.z-V2.z)*(T1.y-T2.y));
       dqdv=-den*((V1.z-V2.z)*(T0.x-T2.x)-(V0.z-V2.z)*(T1.x-T2.x));

       aa=V0.x+(dpdu*P->x)+(dpdv*P->y);
       cc=V0.z+(dqdu*P->x)+(dqdv*P->y);
       bb=-(N.x*aa+N.z*cc+D)/N.y;
      }
      else {                            // projekcia do roviny XY
       if (N.z==0.0f) N.z=1.0f;

       dpdu= den*((V1.x-V2.x)*(T0.y-T2.y)-(V0.x-V2.x)*(T1.y-T2.y));
       dpdv=-den*((V1.x-V2.x)*(T0.x-T2.x)-(V0.x-V2.x)*(T1.x-T2.x));
       dqdu= den*((V1.y-V2.y)*(T0.y-T2.y)-(V0.y-V2.y)*(T1.y-T2.y));
       dqdv=-den*((V1.y-V2.y)*(T0.x-T2.x)-(V0.y-V2.y)*(T1.x-T2.x));

       aa=V0.x+(dpdu*P->x)+(dpdv*P->y);
       bb=V0.y+(dqdu*P->x)+(dqdv*P->y);
       cc=-(N.x*aa+N.y*bb+D)/N.z;
      }

 // 3D suradnice lumelu
 P->x=aa; P->y=bb; P->z=cc;
}
