// hlavickove subory *******************************************************
#include "includes.h"
// vlastny .h subor
#include "avatar.h"

// konstruktor *************************************************************
CAvatar::CAvatar()
{
 reset(); // reset parametrov Avatara
}

// destruktor **************************************************************
CAvatar::~CAvatar()
{

}

// inicializacia Avatara ***************************************************
void CAvatar::init(void)
{
 glMatrixMode(GL_PROJECTION); // vyber projekcnej matice
 glLoadIdentity();            // zresetovanie projekcnej matice

 if (scr_h==0) scr_h=1;
 // nastavenie perspektivneho premietania
 // fov - zorny uhol pohladu, W/H - aspect - pomer sirky a vysky kresliacej plochy
 // viewnear - minimalny dohlad, viewfar - maximalny dohlad
 gluPerspective(fov,(GLfloat) scr_w/(GLfloat) scr_h,viewnear,viewfar);

 // nastavenie rozmerov OpenGL kresliacej plochy
 glViewport(0,0,scr_w,scr_h);

 // TRANSFORMUJEME SCENU, NIE KAMERU!
 glMatrixMode(GL_MODELVIEW); // vyber modelview matice
 glLoadIdentity();           // zresetovanie modelview matice
}

// reset parametrov Avatara ************************************************
void CAvatar::reset(void)
{
 // sposob vypoctu pohybu, 0=chodza, 1=let
 movetype=1;

 // vyska oci Avatara = y-ova suradnica
 avatarEye=1.8f;
 // sirka Avatara
 avatarWidth=1.0f;

 mat.identity(); // nastavenie jednotkovej transformacnej matice Avatara

 actpos.set(0.0f,-avatarEye,0.0f); oldpos=actpos; // aktualna a povodna poloha
 move.set(0.0f,0.0f,0.0f);                        // prirastok pre zmenu polohy
 angle.set(0.0f,0.0f,0.0f);                       // natocenie
 rotate.set(0.0f,0.0f,0.0f);                      // prirastok pre zmenu natocenia
 target.set(0.0f,0.0f,0.0f);                      // vektor smerujuci dopredu

 // koeficienty
 speed=4.0f; slide=4.0f; altit=4.0f; // rychlost, bocny pohyb, zmena vysky
 acceleration=2.0f; gravity=9.81f;   // zrychlenie, gravitacne zrychlenie
 viewnear=0.1f; viewfar=1000.0f;     // minimalny a maximalny dohlad
 fov=60.0f;                          // zorny uhol pohladu
 jumpPower=0.0f;                     // sila skoku

 gravityTime=0.0f; // cas posobenia gravitacie

 // indikatory pre vykonavanie definovanych posunov Avatara
 goForward=0.0f; goBack=0.0f; goUp=0.0f; goDown=0.0f; goRight=0.0f; goLeft=0.0f;
 goJump=0.0f; goFast=0.0f;
 // strelba
 isFire1=0; isFire2=0;
}

// nastavenie prirastkov pre zmenu polohy Avatara **************************
/*
 rychlost  v=s/t [m/s]  draha  s=vt [m]
           v=at [m/s]
 zrychlenie             a=s/t2 [m/s2]
 gravitacne zrychlenie  g=v/t [m/s2], g na Zemi=9.81 m/s2
                        g=s/t2  draha  s=gt2 [m]
*/
void CAvatar::setmovxyz(GLfloat DT)
{
 // ak su stlacene sucasne dve klavesy, Avatar sa nehybe
 // niektore hodnoty su upravene pre dosiahnutie realnejsieho vnemu
 goFast=1.0f+50.0f*goFast;
 move.set(-goFast*(goRight-goLeft)*slide*DT,    // vlavo a vpravo
          -goFast*(goUp-goDown)*altit*DT,       // hore a dole
          -goFast*(goForward-goBack)*speed*DT); // dopredu a dozadu

 // pri type pohybu chodza posobi gravitacia
 if (movetype==0) {
  if (gravityTime==0.0f) { // ak je Avatar na zemi
   if (DT!=0.0f) jumpPower=goJump*sqrt(1.0f/(1000.0f/DT)); // vypocet sily skoku (zavisle od fps!)
  }
  // zvacsovanie sa casu trvania skoku/posobenia gravitacie
  gravityTime+=DT;
  // vypocet klesania v zavislosti od skoku a gravitacie
  move.y=jumpPower-gravity*gravityTime/20.0f;
 }

 // rozdiel poloh AVATARa nemoze byt vacsi ako jeho sirka
 // pretoze nastanu chyby pri vypocte kolizie
 //if ((movx>0.0f)&&(movx>avatarWidth)) movx=avatarWidth;
 //else if (movx<-avatarWidth) movx=-avatarWidth;
 //if ((movy>0.0f)&&(movy>avatarWidth)) movy=avatarWidth;
 //else if (movy<-avatarWidth) movy=-avatarWidth;
 //if ((movz>0.0f)&&(movz>avatarWidth)) movz=avatarWidth;
 //else if (movz<-avatarWidth) movz=-avatarWidth;

 // zrusenie dalsieho mozneho posunu, alebo skoku
 goForward=0.0f; goBack=0.0f; goUp=0.0f; goDown=0.0f; goRight=0.0f; goLeft=0.0f;
 goJump=0.0f; goFast=0.0f;
}

// nastavenie prirastkov pre zmenu natocenia Avatara ***********************
void CAvatar::setrotxy(GLfloat X,GLfloat Y)
{
 // hodnoty su upravene pre dosiahnutie realnejsieho vnemu
 rotate.x=X/10.0f; rotate.y=Y/10.0f;
}

// nastavenie prirastkov pre zmenu natocenia Avatara ***********************
void CAvatar::setrotzl(GLfloat Z)
{
 rotate.z=-Z;
}

// nastavenie prirastkov pre zmenu natocenia Avatara ***********************
void CAvatar::setrotzr(GLfloat Z)
{
 rotate.z+=Z;
}

// aktualizacia parametrov Avatara *****************************************
void CAvatar::update(void)
{
 // ulozenie povodnej polohy
 oldpos=actpos;

 // SPECIALNY PRISTUP K ROTACII A NASLEDNEMU POSUNU KAMERY!
 // aktualna je vzdy modelview matica, ak zanedbame mensie nepresnosti,
 // ktore nastanu z dovodu nepouzitia glLoadIdentity(), nemusime vzdy
 // resetovat modelview maticu, ani ju nasobit transformacnou maticou Avatara

 //glMatrixMode(GL_MODELVIEW); // aktualna je vzdy modelview matica...

 // modelview matica sa nahradi transformacnou maticou Avatara
 // hodnoty prvkov pre rotaciu su inkrementovane
 // hodnoty prvkov pre posun su ponechane na 0.0f
 glLoadMatrixf(mat.get());

 switch (movetype) {
  case 0:  // chodza
   glRotatef(rotate.x,mat.m[0],mat.m[4],mat.m[8]);
   glRotatef(rotate.y,0.0f,1.0f,0.0f);
   // transformacna matica Avatara sa nahradi modelview maticou
   glGetFloatv(GL_MODELVIEW_MATRIX,mat.get());

   // nastavenie novej polohy
   v.set(mat.m[0]*move.x-mat.m[2]*move.z,
         move.y,
         mat.m[8]*move.x-mat.m[10]*move.z);
   actpos+=v;
   // nastavenie noveho natocenia
   angle+=rotate;
  break;
  case 1:  // let
   glRotatef(rotate.x,mat.m[0],mat.m[4],mat.m[8]);
   glRotatef(rotate.y,mat.m[1],mat.m[5],mat.m[9]);
   glRotatef(rotate.z,mat.m[2],mat.m[6],mat.m[10]);
   // transformacna matica Avatara sa nahradi modelview maticou
   glGetFloatv(GL_MODELVIEW_MATRIX,mat.get());

   // nastavenie novej polohy
   v.set(mat.m[0]*move.x+mat.m[1]*move.y-mat.m[2]*move.z,
         mat.m[4]*move.x+mat.m[5]*move.y-mat.m[6]*move.z,
         mat.m[8]*move.x+mat.m[9]*move.y-mat.m[10]*move.z);
   actpos+=v;
   // nastavenie noveho natocenia
   angle+=rotate;
  break;
  default: // dalsie typy pohybu...
  break;
 }

 // zrusenie dalsej moznej rotacie
 rotate.x=0.0f; rotate.y=0.0f; // rotate.z=0.0f;
 // vektor smerujuci dopredu
 target.set(mat.m[2],mat.m[6],mat.m[10]);

 // zresetovanie modelview matice
 //glLoadIdentity(); // nemusi sa vzdy resetovat...

 // TRANSFORMACIA SCENY PODLA NASTAVENI AKTUALNEJ KAMERY!
 // ROTACIA
 //glMultMatrixf(mat.get()); // nemusi sa vzdy nasobit...
 // POSUN
 // ovplyvni modelview maticu, nie transformacnu maticu Avatara
 glTranslatef(actpos.x,actpos.y,actpos.z);
}
