// hlavickove subory *******************************************************
#include "includes.h"
// vlastny .h subor
#include "load3ds.h"

// konstruktor *************************************************************
CLoad3ds::CLoad3ds()
{
 // reset poctu objektov a materialov, reset zoznamov
 model.objNum=0; model.obj.reset();
 model.matNum=0; model.mat.reset();

 tmpObj=NULL;
 tmpMat=NULL;

 buf=NULL;
}

// destruktor **************************************************************
CLoad3ds::~CLoad3ds()
{
 // uvolnenie vyhradenej pamate
 loopk(model.objNum) {
  DELETEA(model.obj[k].verts);     // pole vrcholov
  DELETEA(model.obj[k].normals);   // pole normal
  DELETEA(model.obj[k].faces);     // pole stien
  DELETEA(model.obj[k].texCoords); // pole texturovacich suradnic
 }
}

// nacitanie Chunk-ID a Chunk-length infobloku *****************************
void CLoad3ds::ReadChunkID(Schunk *C)
{
 C->Bread=fread(&C->id,1,2,fp);      // Chunk-ID     - 2 Byte
 C->Bread+=fread(&C->length,1,4,fp); // Chunk-length - 4 Byte

 // INFO: typ a velkost infobloklu
/*
 switch (C->id) {
  case RGB_COLOR:      printf("chunk id: RGB_COLOR, length: %u\n",C->length); break;
  case MAIN:           printf("chunk id: MAIN, length: %u\n",C->length); break;
  case VERSION:        printf("chunk id: VERSION, length: %u\n",C->length); break;
  case EDITOR:         printf("chunk id: EDITOR, length: %u\n",C->length); break;
  case OBJECT:         printf("chunk id: OBJECT, length: %u\n",C->length); break;
  case MESH:           printf("chunk id: MESH, length: %u\n",C->length); break;
  case VERTS_LIST:     printf("chunk id: VERTS_LIST, length: %u\n",C->length); break;
  case FACESDESCR:     printf("chunk id: FACESDESCR, length: %u\n",C->length); break;
  case FACESMAT_LIST:  printf("chunk id: FACESMAT_LIST, length: %u\n",C->length); break;
  case TEXCOORDS_LIST: printf("chunk id: TEXCOORDS_LIST, length: %u\n",C->length); break;
  case MATERIAL:       printf("chunk id: MATERIAL, length: %u\n",C->length); break;
  case MAT_NAME:       printf("chunk id: MAT_NAME, length: %u\n",C->length); break;
  case MAT_AMBIENT:    printf("chunk id: MAT_AMBIENT, length: %u\n",C->length); break;
  case MAT_DIFFUSE:    printf("chunk id: MAT_DIFFUSE, length: %u\n",C->length); break;
  case MAT_SPECULAR:   printf("chunk id: MAT_SPECULAR, length: %u\n",C->length); break;
  case MAT_MAP:        printf("chunk id: MAT_MAP, length: %u\n",C->length); break;
  case MAT_FILENAME:   printf("chunk id: MAT_FILENAME, length: %u\n",C->length); break;
  default:             printf("chunk id: ...%x, length: %u\n",C->id,C->length); break;
 }
*/
}

// nacitanie obsahu 3ds suboru *********************************************
int CLoad3ds::load(const char *fname)
{
 // otvorenie suboru
 fp=fopen(fname,"rb"); // read binary
 if (!fp) { printf("Nemozno otvorit subor: %s!",fname); return(1); }

 // INFO: 3ds subor
 printf("....................................\n");
 printf("3ds subor: %s\n",fname);

 // uvolnenie vyhradenej pamate
 if (model.objNum!=0) {
  loopk(model.objNum) {
   DELETEA(model.obj[k].verts);     // pole vrcholov
   DELETEA(model.obj[k].normals);   // pole normal
   DELETEA(model.obj[k].faces);     // pole stien
   DELETEA(model.obj[k].texCoords); // pole texturovacich suradnic
  }
 }

 // reset poctu objektov a materialov, reset zoznamov
 model.objNum=0; model.obj.reset();
 model.matNum=0; model.mat.reset();

 Schunk firstChunk; // pre nacitanie obsahu infoblokov

 // nacitanie typu 1. infobloku - mal by to byt MAIN
 ReadChunkID(&firstChunk);
 if (firstChunk.id!=MAIN) { printf("Nemozno nacitat MAIN infoblok!"); return(1); }

 // nacitanie obsahu dalsich infoblokov - REKURZIVNA FUNKCIA!
 ReadChunk(&firstChunk);

 // zatvorenie suboru
 fclose(fp);

 // vlastny vypocet normal, pretoze 3ds subor neobsahuje informacie o normalach
 // sucasne sa prehodia suradnice Y a Z pretoze 3ds subor ich ma ulozene opacne
 // ako OpenGL
 ComputeNormals();

 // INFO: nacitanie textur
 printf("textury:\n");

 unsigned int tn=2; // 0 a 1 su pouzite

 loopk(model.matNum) {
  if (strlen(model.mat[k].mapFile)>0) { // ak ma material texturu
   strcpy(tmpName,"data/textures/");
   strcat(tmpName,model.mat[k].mapFile);

   if (texture.load2D(tid[tn],tmpName,T_MIPMAP|T_LINEAR)) {
    printf("Chyba nahratia textury: %s, pouzije sa standardna!\n",tmpName);
    model.mat[k].mapTid=1;
   }
   else model.mat[k].mapTid=tn++; // ciselne oznacenie textury
  }
  else model.mat[k].mapTid=0; // 0-material nema texturu
 }

 // INFO: 3ds subor
 printf("....................................\n");

 return(0);
}

// nacitanie obsahu infoblokov - REKURZIVNA FUNKCIA! ***********************
void CLoad3ds::ReadChunk(Schunk *C)
{
 Schunk tmpChunk; // pre nacitanie obsahu infoblokov

 // nacitanie obsahu infobloku a jeho podblokov az pokial sa nenacita pocet
 // Byte rovnajuci sa celkovej velkosti infobloku
 while (C->Bread<C->length) {
  // nacitanie typu infobloku
  ReadChunkID(&tmpChunk);

  switch (tmpChunk.id) {
   case VERSION:
    tmpChunk.Bread+=fread(&ver,1,4,fp); // verzia - 4 Byte
    printf("verzia: %u\n",ver);
   break;
   case EDITOR:
    // rekurzivne nacitanie podblokov OBJECT a MATERIAL
    ReadChunk(&tmpChunk);
   break;
   case OBJECT:
    // zvacsenie poctu objektov
    model.objNum++; objn=model.objNum-1;
    // pridanie noveho objektu do zoznamu
    model.obj.add();
    // nacitanie nazvu objektu - 1 Byte / prvok
    i=0; tmpChunk.Bread+=fread((model.obj[objn].name+i),1,1,fp);
    while (*(model.obj[objn].name+i)!=0) {
     i++; tmpChunk.Bread+=fread((model.obj[objn].name+i),1,1,fp);
    }
    printf("nazov objektu: %s\n",model.obj[objn].name);
    // nacitanie podbloku OBJECT
    ReadObjectChunk(&tmpChunk);
   break;
   case MATERIAL:
    // zvacsenie poctu materialov
    model.matNum++; matn=model.matNum-1;
    // pridanie noveho materialu do zoznamu
    model.mat.add();
    // nacitanie podbloku MATERIAL
    ReadMaterialChunk(&tmpChunk);
   break;
   default:
    // preskocenie ostatnych infoblokov
    buf=new unsigned int[tmpChunk.length-tmpChunk.Bread];
    if (buf==NULL) { printf("Nemozno vyhradit pamat!\n"); }
    tmpChunk.Bread+=fread(buf,1,tmpChunk.length-tmpChunk.Bread,fp);
    DELETEA(buf);
   break;
  }

  // pocet nacitanych Bytov
  C->Bread+=tmpChunk.Bread;
  // INFO: pocet nacitanych Bytov z celkoveho poctu
  //printf("Bread: %u/%u\n",C->Bread,C->length);
 }
}

// nacitanie infobloku OBJECT - REKURZIVNA FUNKCIA! ************************
void CLoad3ds::ReadObjectChunk(Schunk *C)
{
 Schunk tmpChunk; // pre nacitanie obsahu infoblokov

 // nacitanie obsahu infobloku a jeho podblokov az pokial sa nenacita pocet
 // Byte rovnajuci sa celkovej velkosti infobloku
 while (C->Bread<C->length) {
  // nacitanie typu infobloku
  ReadChunkID(&tmpChunk);

  switch (tmpChunk.id) {
   case MESH:
    // rekurzivne nacitanie podblokov VERTS_LIST, FACESDESCR, FACESMAT_LIST, TEXCOORDS_LIST
    ReadObjectChunk(&tmpChunk);
   break;
   case VERTS_LIST:
    // nacitanie poctu vrcholov - 2 Byte
    tmpChunk.Bread+=fread(&(model.obj[objn].vertsNum),1,2,fp);
    // INFO: pocet vrcholov objektu
    //fprintf(stderr,"pocet vrcholov: %u\n",model.obj[objn].vertsNum);
    // priprava pola vrcholov a normal
    model.obj[objn].verts=new vec3[model.obj[objn].vertsNum];
    model.obj[objn].normals=new vec3[model.obj[objn].vertsNum];
    if ((model.obj[objn].verts==NULL)||
        (model.obj[objn].normals==NULL)) { printf("Nemozno vyhradit pamat!\n"); }
    // nacitanie suradnic vrcholov - float x 3
    tmpChunk.Bread+=fread(model.obj[objn].verts,1,tmpChunk.length-tmpChunk.Bread,fp);
    // INFO: vrcholy
    //loopk(model.obj[objn].vertsNum) {
    // printf("%f %f %f\n",model.obj[objn].verts[k].x,
    //                     model.obj[objn].verts[k].z,
    //                     model.obj[objn].verts[k].y);
    //}
   break;
   case FACESDESCR:
    // nacitanie poctu stien - 2 Byte
    tmpChunk.Bread+=fread(&model.obj[objn].facesNum,1,2,fp);
    // INFO: pocet stien objektu
    //fprintf(stderr,"pocet stien: %u\n",model.obj[objn].facesNum);
    // priprava pola stien
    model.obj[objn].faces=new Sindices[model.obj[objn].facesNum];
    if (model.obj[objn].faces==NULL) { printf("Nemozno vyhradit pamat!\n"); }
    // nacitanie indexov vrcholov do pola stien - 2 Byte x 4
    loopk(model.obj[objn].facesNum) {
     tmpChunk.Bread+=fread(&(model.obj[objn].faces[k]),1,2*4,fp);
    }
   break;
   case FACESMAT_LIST:
    // nacitanie nazvu materialu - 1 Byte / prvok
    i=0; tmpChunk.Bread+=fread((tmpName+i),1,1,fp);
    while (*(tmpName+i)!=0) {
     i++; tmpChunk.Bread+=fread((tmpName+i),1,1,fp);
    }
    // ziskanie indexu materialu zo zoznamu materialov
    loopk(model.matNum) {
     if (strcmp(tmpName,model.mat[k].name)==0) {
      model.obj[objn].matIndex=k;
      // INFO: index do zoznamu materialov
      //printf("index do zoznamu materialov: %i\n",k);
      break; // ukoncenie loopk
     }
     else {
      model.obj[objn].matIndex=-1;
      // INFO: index do zoznamu materialov
      //printf("index do zoznamu materialov: %i\n",k);
     }
    }
    // preskocenie dalsich informacii ulozenych v aktualnom infobloku
    buf=new unsigned int[tmpChunk.length-tmpChunk.Bread];
    if (buf==NULL) { printf("Nemozno vyhradit pamat!\n"); }
    tmpChunk.Bread+=fread(buf,1,tmpChunk.length-tmpChunk.Bread,fp);
    DELETEA(buf);
   break;
   case TEXCOORDS_LIST:
    // pocet texturovacich suradnic - 2 Byte
    tmpChunk.Bread+=fread(&model.obj[objn].texcoordsNum,1,2,fp);
    // INFO: pocet texturovacich suradnic
    //fprintf(stderr,"pocet texturovacich suradnic: %u\n",model.obj[objn].texcoordsNum);
    // priprava pola teturovacich suradnic
    model.obj[objn].texCoords=new vec2[model.obj[objn].texcoordsNum];
    if (model.obj[objn].texCoords==NULL) { printf("Nemozno vyhradit pamat!\n"); }
    // nacitanie texturovacich suradnic - float x 2
    tmpChunk.Bread+=fread(model.obj[objn].texCoords,1,tmpChunk.length-tmpChunk.Bread,fp);
   break;
   default:
    // preskocenie ostatnych infoblokov
    buf=new unsigned int[tmpChunk.length-tmpChunk.Bread];
    if (buf==NULL) { printf("Nemozno vyhradit pamat!\n"); }
    tmpChunk.Bread+=fread(buf,1,tmpChunk.length-tmpChunk.Bread,fp);
    DELETEA(buf);
   break;
  }

  // pocet nacitanych Bytov
  C->Bread+=tmpChunk.Bread;
  // INFO: pocet nacitanych Bytov z celkoveho poctu
  //printf("Bread: %u/%u\n",C->Bread,C->length);
 }
}

// nacitanie infobloku MATERIAL ********************************************
void CLoad3ds::ReadMaterialChunk(Schunk *C)
{
 Schunk tmpChunk; // pre nacitanie obsahu infoblokov

 // nacitanie obsahu infobloku a jeho podblokov az pokial sa nenacita pocet
 // Byte rovnajuci sa celkovej velkosti infobloku
 while (C->Bread<C->length) {
  // nacitanie typu infobloku
  ReadChunkID(&tmpChunk);

  switch (tmpChunk.id) {
   case MAT_NAME:
    // nazov materialu - 1 Byte / prvok
    tmpChunk.Bread+=fread(model.mat[matn].name,1,tmpChunk.length-tmpChunk.Bread,fp);
    // INFO: nazov materialu
    //printf("nazov materialu: %s\n",model.mat[matn].name);
   break;
   case MAT_DIFFUSE:
    // farba je ulozena v dalsom podbloku
    ReadChunkID(&ch);
    // nacitanie difuznej zlozky farby (R,G,B) - 1 Byte x 3
    ch.Bread+=fread(model.mat[matn].difColor,1,1*3,fp);
    // zvysenie poctu nacitanych Bytov
    tmpChunk.Bread+=ch.Bread;
   break;
   case MAT_MAP:
    // nazov suboru s texturou je ulozeny v dalsom podbloku
    ReadChunkID(&ch);
    // nacitanie nazvu suboru s texturou - 1 Byte / prvok
    ch.Bread+=fread(model.mat[matn].mapFile,1,ch.length-ch.Bread,fp);
    // INFO: nazov suboru s texturou
    //printf("nazov suboru s texturou: %s\n",model.mat[matn].mapFile);
    // zvysenie poctu nacitanych Bytov
    tmpChunk.Bread+=ch.Bread;
   break;
   default:
   // preskocenie ostatnych infoblokov
   buf=new unsigned int[tmpChunk.length-tmpChunk.Bread];
   if (buf==NULL) { printf("Nemozno vyhradit pamat!\n"); }
   tmpChunk.Bread+=fread(buf,1,tmpChunk.length-tmpChunk.Bread,fp);
   DELETEA(buf);
   break;
  }

  // pocet nacitanych Bytov
  C->Bread+=tmpChunk.Bread;
  // INFO: pocet nacitanych Bytov z celkoveho poctu
  //printf("Bread: %u/%u\n",C->Bread,C->length);
 }
}

// vypocet normal **********************************************************
int CLoad3ds::ComputeNormals(void)
{
 vec3 v[3];    // vrcholy steny (trojuholnika)
 vec3 *n=NULL; // pole pre docasne ulozenie normal stien
 vec3 a,b,c;   // pomocne vektory
 float s;      // pomocna premenna

 // ak neboli nacitane objekty
 if (model.objNum==0) return(1);

 // pre vsetky objekty
 loopk(model.objNum) {
  tmpObj=&(model.obj[k]); // aktualny objekt

  // prehodenie Y a Z suradnic
  loopi(tmpObj->vertsNum) {
   s=tmpObj->verts[i].y;
   tmpObj->verts[i].y=tmpObj->verts[i].z;
   tmpObj->verts[i].z=-s;
  }

  // priprava pola pre docasne ulozenie normal stien
  n=new vec3[tmpObj->facesNum];
  if (n==NULL) { printf("Nemozno vyhradit pamat!\n"); return(1); }

  // vypocet normal pre vsetky steny
  loopi(tmpObj->facesNum) {
   // suradnice vrcholov steny
   v[0]=tmpObj->verts[tmpObj->faces[i].ind[0]];
   v[1]=tmpObj->verts[tmpObj->faces[i].ind[1]];
   v[2]=tmpObj->verts[tmpObj->faces[i].ind[2]];

   // dva vektory
   a=v[1]-v[0]; b=v[2]-v[0];
   // vektorovy sucin dvoch vektorov = normala steny
   n[i].cross(a,b);
  }

  // vypocet normal pre vsetky vrcholy
  loopj(tmpObj->vertsNum) { // pre vsetky vrcholy

   c.set(0.0f,0.0f,0.0f); // pre scitavanie vektorov
   s=0.0f;                // pocet zdielanych vrcholov
   loopi(tmpObj->facesNum) { // pre vsetky steny
    if ((tmpObj->faces[i].ind[0]==j)||(tmpObj->faces[i].ind[1]==j)||(tmpObj->faces[i].ind[2]==j)) {
     // ak dana stena zdiela vrchol s inou stenou, ich normaly sa scitaju
     // ziskame tak hladky prechod farieb medzi dvoma spojenymi stenami
     c.add(n[i]);
     s+=1.0f;
    }
   }

   // konecny vypocet normaly vrcholu:
   // normala = sucet normal stien, ktore zdielaju vrchol / pocet takychto stien
   tmpObj->normals[j]=c.div(s);
   // normalizacia normaly
   tmpObj->normals[j].normalize();

   // INFO: normala vrcholu
   //printf("vrchol: %u, normala: %f,%f,%f\n",j,tmpObj->normals[j].x,
   //                                           tmpObj->normals[j].y,
   //                                           tmpObj->normals[j].z);
  }

  // uvolnenie vyhradenej pamate
  DELETEA(n);
 }

 return(0);
}

// vykreslenie 3ds modelu **************************************************
int CLoad3ds::forcedraw(void)
{
 // ak neboli nacitane objekty
 if (model.objNum==0) return(1);

 // ulozenie potrebnych atributov
 glPushAttrib(GL_TEXTURE_BIT);

 // pre vsetky objekty
 loopk(model.objNum) {
  tmpObj=&(model.obj[k]); // aktualny objekt

  isTex=false;
  if (tmpObj->matIndex!=-1) { // ak ma objekt material
   tmpMat=&(model.mat[tmpObj->matIndex]); // aktualny material

   if (tmpMat->mapTid!=0) { // ak ma material texturu
    glEnable(GL_TEXTURE_2D);
    // textura
    glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[tmpMat->mapTid]]);
    isTex=true;
   }
   else { // material nema texturu - iba farbu
    glDisable(GL_TEXTURE_2D);
    // farba
    glColor4f((GLfloat) tmpMat->difColor[0]/255.0f,
              (GLfloat) tmpMat->difColor[1]/255.0f,
              (GLfloat) tmpMat->difColor[2]/255.0f,1.0f);
   }
  }

  glBegin(GL_TRIANGLES);
   // pre vsetky steny
   loopj(tmpObj->facesNum) {
    // pre vsetky vrcholy
    loopi(3) {
     l=tmpObj->faces[j].ind[i];
     // normala
     glNormal3f(tmpObj->normals[l].x,tmpObj->normals[l].y,tmpObj->normals[l].z);
     if (isTex) { // objekt ma material, ktory ma texturu
      // texturovacie suradnice
      glTexCoord2f(tmpObj->texCoords[l].x,tmpObj->texCoords[l].y);
     }
     // suradnice vrcholu
     glVertex3f(tmpObj->verts[l].x,tmpObj->verts[l].y,tmpObj->verts[l].z);
    }
   }
  glEnd();
 }

 // nastavenie standardnej farby - biela
 glColor4f(1.0f,1.0f,1.0f,1.0f);

 glPopAttrib(); // vyber ulozenych atributov - TEXTURE

 return(0);
}

// ziskanie trojuholnikov patriacich objektom ******************************
// ziskanie informacii o svetelnych zdrojoch definovanych v 3ds subore
int CLoad3ds::get_tris_and_lights(void)
{
 GLuint trisnum;     // pocet pridanych trojuholnikov
 vec4 triscolor;     // farba trojuholnikov konkretneho objektu
 GLuint tristexture; // identifikator OpenGL textury konkretneho objektu

 GLuint lightsnum; // pocet pridanych svetelnych zdrojov
 vec4 lightcolor;  // farba svetelnych zdrojov

 // ak neboli nacitane objekty
 if (model.objNum==0) return(1);

 trisnum=0; triscolor.set(1.0f,1.0f,1.0f,1.0f); tristexture=0;
 // reset zoznamu vsetkych trojuholnikov
 tris.reset();

 lightsnum=0; lightcolor.set(1.0f,1.0f,1.0f,1.0f);
 // reset zoznamu vsetkych svetelnych zdrojov
 lights.reset();

 // pre vsetky objekty
 loopk(model.objNum) {
  tmpObj=&(model.obj[k]); // aktualny objekt
  // ak je objektom svetelny zdroj
  if (strstr(tmpObj->name,"light")!=NULL) {
   if (tmpObj->matIndex!=-1) { // ak ma objekt material
    tmpMat=&(model.mat[tmpObj->matIndex]); // aktualny material

    // material nema texturu - iba farbu
    lightcolor.x=(GLfloat) tmpMat->difColor[0]/255.0f;
    lightcolor.y=(GLfloat) tmpMat->difColor[1]/255.0f;
    lightcolor.z=(GLfloat) tmpMat->difColor[2]/255.0f;
    lightcolor.w=1.0f;
   }

   lights.add(); // pridanie noveho svetelneho zdroja do zoznamu
   lights[lightsnum].id=lightsnum; // identifikator svetelneho zdroja
   lights[lightsnum].c=lightcolor; // farba
   l=tmpObj->faces[0].ind[0];
   // poloha
   lights[lightsnum].p=tmpObj->verts[l];

   lightsnum++; // zvysenie poctu pridanych svetelnych zdrojov
  }
  // ak objekt nie je svetelny zdroj, ide o bezny 3D objekt
  else {
   isTex=false;
   if (tmpObj->matIndex!=-1) { // ak ma objekt material
    tmpMat=&(model.mat[tmpObj->matIndex]); // aktualny material

    if (tmpMat->mapTid!=0) { // ak ma material texturu
     tristexture=texture.tex_id[tid[tmpMat->mapTid]];
     isTex=true;
    }
    else { // material nema texturu - iba farbu
     triscolor.x=(GLfloat) tmpMat->difColor[0]/255.0f;
     triscolor.y=(GLfloat) tmpMat->difColor[1]/255.0f;
     triscolor.z=(GLfloat) tmpMat->difColor[2]/255.0f;
     triscolor.w=1.0f;
    }
   }

   // pre vsetky steny
   loopj(tmpObj->facesNum) {

    tris.add(); // pridanie noveho trojuholnika do zoznamu
    tris[trisnum].id=trisnum;  // identifikator trojuholnika
    tris[trisnum].c=triscolor; // farba

    // pre vsetky vrcholy
    loopi(3) {
     l=tmpObj->faces[j].ind[i];
     // normaly vo vrcholoch
     tris[trisnum].vn[i]=tmpObj->normals[l];
     if (isTex) { // objekt ma material, ktory ma texturu
      // texturovacie suradnice
      tris[trisnum].t[i]=tmpObj->texCoords[l];
      // identifikator 1. OpenGL textury
      tris[trisnum].tex1id=tristexture;
      // identifikator 2. OpenGL textury
      tris[trisnum].tex2id=0;
     }
     // suradnice vrcholov
     tris[trisnum].v[i]=tmpObj->verts[l];
    }

    trisnum++; // zvysenie poctu pridanych trojuholnikov
   }
  }
 }

 return(0);
}
